# -*- coding: utf-8 -*-
"""
Created on Wed Nov 22 20:10:30 2017

@author: Samuel
"""

"""Linear Regression 2 - Exercises"""

# In the exercises below we will investigate how robust our models are from 
# exercise 1. 

"""
Exercise 1. Read in the data and split into lower, middle, and high income
level assigning a new variable 'IncomeLevel' to 1, 2, and 3 respectively. Run
simple linear regression models for predicting total revenue spent from 
distance to arena for each of the three income levels. Plot the distribution of
residuals and compute the variance of the residuals for each. What do these 
tell you about our 3 models?

"""
# Read in packages and data

import pandas as pd
from pandas.tools.plotting import scatter_matrix
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score, mean_squared_error
from sklearn.feature_selection import RFE
from sklearn import preprocessing
import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm
from statsmodels.formula.api import ols
import seaborn.apionly as sns
from scipy.stats import f

# Data
xls = pd.ExcelFile('C:/Users/Samuel/Downloads/Business-Track-Application-Datasets.xlsx')
data = xls.parse('Training Data')
del data['Unnamed: 0']

# Create variable IncomeLevel
data.loc[data['Income']<=50000, 'IncomeLevel'] = 1
data.loc[(data['Income']>50000) & (data['Income']<=125000), 'IncomeLevel'] = 2 
data.loc[data['Income']>125000, 'IncomeLevel'] = 3         

X_dist1 = data[data['IncomeLevel']==1]['DistToArena']
X_dist2 = data[data['IncomeLevel']==2]['DistToArena']
X_dist3 = data[data['IncomeLevel']==3]['DistToArena']

y_rev1 = data[data['IncomeLevel']==1]['TotRevSpend']
y_rev2 = data[data['IncomeLevel']==2]['TotRevSpend']
y_rev3 = data[data['IncomeLevel']==3]['TotRevSpend']

# Create SLR models for predicting TotRevSpend
slr1_dist_rev = ols(formula='y_rev1 ~ X_dist1', 
                    data=data[data['IncomeLevel']==1]).fit()
slr2_dist_rev = ols(formula='y_rev2 ~ X_dist2', 
                    data=data[data['IncomeLevel']==2]).fit()
slr3_dist_rev = ols(formula='y_rev3 ~ X_dist3', 
                    data=data[data['IncomeLevel']==3]).fit()

# Get residuals 
slr1_dist_rev.resid
slr2_dist_rev.resid
slr3_dist_rev.resid

# Compute the variance of the residuals
slr1_dist_rev.resid.var(ddof=1) # 36970.438
slr2_dist_rev.resid.var(ddof=1) # 57237.164
slr3_dist_rev.resid.var(ddof=1) # 56949.223
## While income level 1 has significantly less variance, we should keep in mind
## that it also had only about 150 data points whereas the other two income
## levels both had at least 400 data points. 
                     
# Plot a histogram of each of the residuals
fig = plt.figure()

ax1 = fig.add_subplot(131)
ax1.hist(slr1_dist_rev.resid, bins=20, ec='black')
ax1.set_title('Residuals for Income Level 1', fontsize=24)
# Looks a little right skewed

ax2 = fig.add_subplot(132)
ax2.hist(slr2_dist_rev.resid, bins=20, ec='black')
ax2.set_title('Residuals for Income Level 2', fontsize=24)
# Looks more like a normal distribution

ax3 = fig.add_subplot(133)
ax3.hist(slr3_dist_rev.resid, bins=20, ec='black')
ax3.set_title('Residuals for Income Level 3', fontsize=24)
# Looks somewhat like a normal distribution, perhaps even bi-modal

## Will have to perform further testing of the residuals

"""
Exercise 2. Test for normality of our three models by creating Q-Q plots.
Also overlay scatter plots with our SLR models with 90% confidence intervals.
You can change the amount of confidence you want in sns.lmplot with the 
parameter "ci" set to an integer between 0 and 100. Plot the residuals vs. 
fitted values of our models. What does these visualizations tell you regarding 
the assumptions of linear regression.
"""
sm.qqplot(slr1_dist_rev.resid, line='s')
plt.title('Q-Q Plot to Test Normality of Residuals for Income Level 1', fontsize=24)
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)
## Thinner lower tails

sm.qqplot(slr2_dist_rev.resid, line='s')
plt.title('Q-Q Plot to Test Normality of Residuals for Income Level 2', fontsize=24)
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)
## Thinner tails than normal distribution

sm.qqplot(slr3_dist_rev.resid, line='s')
plt.title('Q-Q Plot to Test Normality of Residuals for Income Level 3', fontsize=24)
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)
## Thinner tails than normal distribution

# Best fit lines with confidence intervals

# Income Level 1
sns.lmplot(x='DistToArena', y='TotRevSpend', data=data[data['IncomeLevel']==1],
           line_kws={'color': 'red'}, scatter_kws={'color': 'black'}, ci=90)
plt.xlabel('Distance to Arena (miles)', fontsize=18)
plt.ylabel('Total Revenue Spent ($)', fontsize=18)

# Income Level 2
sns.lmplot(x='DistToArena', y='TotRevSpend', data=data[data['IncomeLevel']==2],
           line_kws={'color': 'red'}, scatter_kws={'color': 'black'}, ci=90)
plt.xlabel('Distance to Arena (miles)', fontsize=18)
plt.ylabel('Total Revenue Spent ($)', fontsize=18)

# Income Level 3
sns.lmplot(x='DistToArena', y='TotRevSpend', data=data[data['IncomeLevel']==3],
           line_kws={'color': 'red'}, scatter_kws={'color': 'black'}, ci=90)
plt.xlabel('Distance to Arena (miles)', fontsize=18)
plt.ylabel('Total Revenue Spent ($)', fontsize=18)

# Get fitted values of each of our models
slr1_dist_rev_fitted = slr1_dist_rev.fittedvalues
slr2_dist_rev_fitted = slr2_dist_rev.fittedvalues
slr3_dist_rev_fitted = slr3_dist_rev.fittedvalues

# Plot the residuals vs fitted values for each
fig = plt.figure()

ax1 = fig.add_subplot(131)
ax1.scatter(slr1_dist_rev_fitted, slr1_dist_rev.resid, color='black')
ax1.plot(slr1_dist_rev_fitted, np.zeros((len(slr1_dist_rev_fitted),)))
ax1.set_title('Residuals vs. Fitted Values for Income Level 1', fontsize=24)
ax1.set_xlabel('Fitted Values', fontsize=18)
ax1.set_ylabel('Residuals', fontsize=18)

ax2 = fig.add_subplot(132)
ax2.scatter(slr2_dist_rev_fitted, slr2_dist_rev.resid, color='black')
ax2.plot(slr2_dist_rev_fitted, np.zeros((len(slr2_dist_rev_fitted),)))
ax2.set_title('Residuals vs. Fitted Values for Income Level 2', fontsize=24)
ax2.set_xlabel('Fitted Values', fontsize=18)
ax2.set_ylabel('Residuals', fontsize=18)

ax3 = fig.add_subplot(133)
ax3.scatter(slr3_dist_rev_fitted, slr3_dist_rev.resid, color='black')
ax3.plot(slr3_dist_rev_fitted, np.zeros((len(slr3_dist_rev_fitted),)))
ax3.set_title('Residuals vs. Fitted Values for Income Level 3', fontsize=24)
ax3.set_xlabel('Fitted Values', fontsize=18)
ax3.set_ylabel('Residuals', fontsize=18)

## Here again, we're seeing this diamond shaped residual plots. The is 
## probably due to the fact that this data seems to be censored i.e. there
## seems to be clusters of points at maximum ($1000) and minimum ($0) values 
## of total revenue spent. The residual plot for income level 1 looks less so 
## perhaps because there are fewer data points but also since these fans
## have lower income, they're probably less likely to be spending at least 
## $1000 in the NBA season. It seems that our middle and high incomes fans
## have non-constant variance but we will test this with the Breusch-Pagan 
## test as well for practice. 

"""
Exercise 3. Test formally for heteroscedasticity using the Breusch-Pagan test.
What are the conclusions of this? Also compute the influence of each point
for all three models with Cook's distance. Are there any patterns with these
potential outliers? Considering coloring these points differently on scatter
plots. 
"""
import statsmodels.stats.api as sms
from statsmodels.compat import lzip

name = ['Lagrange multiplier statistic', 'p-value',
        'f-value', 'f p-value']

test_slr1 = sms.het_breuschpagan(slr1_dist_rev.resid, slr1_dist_rev.model.exog)
lzip(name, test_slr1)
# Even though the residual plot looked the least heteroscedastic, the small
# p-value indicates that there is heteroscedasticity.

test_slr2 = sms.het_breuschpagan(slr2_dist_rev.resid, slr2_dist_rev.model.exog)
lzip(name, test_slr2)
# While the residual plot showed a clear pattern indicating heteroscedasticity,
# this p-value is not low enough to reject the notion that there is no
# heteroscedasticity.

test_slr3 = sms.het_breuschpagan(slr3_dist_rev.resid, slr3_dist_rev.model.exog)
lzip(name, test_slr3)
# While the residual plot showed a clear pattern indicating heteroscedasticity,
# this p-value is very high indicating that we cannot reject the notion that 
# there is no heteroscedasticity.

# Outlier Analysis
slr1_dist_rev_influence = slr1_dist_rev.get_influence()
slr2_dist_rev_influence = slr2_dist_rev.get_influence()
slr3_dist_rev_influence = slr3_dist_rev.get_influence()

(slr1_CD, p) = slr1_dist_rev_influence.cooks_distance
(slr2_CD, p) = slr2_dist_rev_influence.cooks_distance
(slr3_CD, p) = slr3_dist_rev_influence.cooks_distance

fig = plt.figure()

ax1 = fig.add_subplot(131)
ax1.stem(np.arange(len(slr1_CD)), slr1_CD)
ax1.set_title('Cook Distance for Income Level 1', fontsize=24)
ax1.grid(.25)

ax2 = fig.add_subplot(132)
ax2.stem(np.arange(len(slr2_CD)), slr2_CD)
ax2.set_title('Cook Distance for Income Level 2', fontsize=24)
ax2.grid(.25)

ax3 = fig.add_subplot(133)
ax3.stem(np.arange(len(slr3_CD)), slr3_CD)
ax3.set_title('Cook Distance for Income Level 3', fontsize=24)
ax3.grid(.25)

# Extract potential outliers with threshold 4 / len(data points)
slr1_CD_select = slr1_CD[slr1_CD > 4 / len(slr1_CD)]
slr2_CD_select = slr2_CD[slr2_CD > 4 / len(slr2_CD)]
slr3_CD_select = slr3_CD[slr3_CD > 4 / len(slr3_CD)]

data1_influence = data[data['IncomeLevel']==1][slr1_CD > 4 / len(slr1_CD)]
data2_influence = data[data['IncomeLevel']==2][slr2_CD > 4 / len(slr2_CD)]
data3_influence = data[data['IncomeLevel']==3][slr3_CD > 4 / len(slr3_CD)]


sns.lmplot(x='DistToArena', y='TotRevSpend', 
           data=data[data['IncomeLevel']==1],
           line_kws={'color': 'orange'}, 
           scatter_kws={'color': 'black'})
plt.scatter(data1_influence['DistToArena'], data1_influence['TotRevSpend'], color='red')
plt.xlabel('Distance to Arena (miles)', fontsize=18)
plt.ylabel('Total Revenue Spent ($)', fontsize=18)
## Outliers look like those who live relatively close to the arena.

sns.lmplot(x='DistToArena', y='TotRevSpend', 
           data=data[data['IncomeLevel']==2],
           line_kws={'color': 'orange'}, 
           scatter_kws={'color': 'black'})
plt.scatter(data2_influence['DistToArena'], data2_influence['TotRevSpend'], color='red')
plt.xlabel('Distance to Arena (miles)', fontsize=18)
plt.ylabel('Total Revenue Spent ($)', fontsize=18)
## Outliers look like those who live relatively close to the arena but 
## don't spend as much money.

sns.lmplot(x='DistToArena', y='TotRevSpend', 
           data=data[data['IncomeLevel']==3],
           line_kws={'color': 'orange'}, 
           scatter_kws={'color': 'black'})
plt.scatter(data3_influence['DistToArena'], data3_influence['TotRevSpend'], color='red')
plt.xlabel('Distance to Arena (miles)', fontsize=18)
plt.ylabel('Total Revenue Spent ($)', fontsize=18)
## Outliers look like those who live relatively close to the arena but 
## but don't spend as much money.

"""
Exercise 4. Let's add the variables GamesWatched and FanComplaints to our 
models. Test for heteroscedasticity, normality of residuals, and 
multicollinearity.
"""
X1 = data[data['IncomeLevel']==1][['DistToArena', 'GamesWatched', 'FanComplaints']]
X2 = data[data['IncomeLevel']==2][['DistToArena', 'GamesWatched', 'FanComplaints']]
X3 = data[data['IncomeLevel']==3][['DistToArena', 'GamesWatched', 'FanComplaints']]

mlr1 = ols('y_rev1 ~ X1', data=data[data['IncomeLevel']==1]).fit()
mlr2 = ols('y_rev2 ~ X2', data=data[data['IncomeLevel']==2]).fit()
mlr3 = ols('y_rev3 ~ X3', data=data[data['IncomeLevel']==3]).fit()

# Regression output
mlr1.summary()
mlr2.summary()
mlr3.summary()

# Normality
sm.qqplot(mlr1.resid, line='s')
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)

sm.qqplot(mlr2.resid, line='s')
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)

sm.qqplot(mlr3.resid, line='s')
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)

## All seem realtively normal except in the tails.

# Heteroscedasticity
fig = plt.figure()

ax1 = fig.add_subplot(131)
ax1.scatter(mlr1.fittedvalues, mlr1.resid, color='black')
plt.plot(mlr1.fittedvalues, np.zeros((len(mlr1.fittedvalues),)), color='orange')
ax1.set_title('Income Level 1', fontsize=24)
ax1.set_xlabel('Fitted Values', fontsize=18)
ax1.set_ylabel('Residual', fontsize=18)

ax2 = fig.add_subplot(132)
ax2.scatter(mlr2.fittedvalues, mlr2.resid, color='black')
plt.plot(mlr2.fittedvalues, np.zeros((len(mlr2.fittedvalues),)), color='orange')
ax2.set_title('Income Level 2', fontsize=24)
ax2.set_xlabel('Fitted Values', fontsize=18)
ax2.set_ylabel('Residual', fontsize=18)

ax3 = fig.add_subplot(133)
ax3.scatter(mlr3.fittedvalues, mlr3.resid, color='black')
plt.plot(mlr3.fittedvalues, np.zeros((len(mlr3.fittedvalues),)), color='orange')
ax3.set_title('Income Level 3', fontsize=24)
ax3.set_xlabel('Fitted Values', fontsize=18)
ax3.set_ylabel('Residual', fontsize=18)

## Same story as before. Residuals look highly dependent on fitted values
## due to censoring of the data.

test_mlr1 = sms.het_breuschpagan(mlr1.resid, mlr1.model.exog)
lzip(name, test_mlr1)
## Low p-value indicates existence of heteroscedasticity

test_mlr2 = sms.het_breuschpagan(mlr2.resid, mlr2.model.exog)
lzip(name, test_mlr2)
## Fairly high p-value does not indicate existence of heteroscedasticity, 
## but residual plot shows there is.

test_mlr3 = sms.het_breuschpagan(mlr3.resid, mlr3.model.exog)
lzip(name, test_mlr3)
## ## Fairly high p-value does not indicate existence of heteroscedasticity, 
## but residual plot shows there is.

# Multicollinearity
from statsmodels.stats.outliers_influence import variance_inflation_factor

# First examine our correlations
X1.corr()
X2.corr()
X3.corr()

# Variance Inflation Factors
X1_multicollinearity = [variance_inflation_factor(X1.values, j)
                        for j in range(X1.values.shape[1])]
print(X1_multicollinearity)

X2_multicollinearity = [variance_inflation_factor(X2.values, j)
                        for j in range(X2.values.shape[1])]
print(X2_multicollinearity)

X3_multicollinearity = [variance_inflation_factor(X3.values, j)
                        for j in range(X3.values.shape[1])]
print(X3_multicollinearity)

## None of these models have severe multicollinearity. To further decrease
## multicollinearity we could standardize our data.

# Can we trust these models?
## Since our residual variance does seem to be non-constant, we can't really
## trust our statistical significance results, but our middle and high
## income fan models do still have an R^2 above 40%. While this isn't great
## it's a good point to continue doing analysis. We may still get some 
## useful predictions from these models. 
"""
Exercise 5. For each of our three income levels, randomly split the data
using "train_test_split" function from the sklearn.model_selection module.
Make the test size 20%. Refit our models using the other 80% of the data.
Predict the total revenue for the last 20%. To see how well our predictions
do, compute the mean squared error and R^2 between our predictions and the 
actual total revenue spent values in the test set. You can do this in either 
scikit-learn or statsmodels.

Note our train_test_split should look like:
    x_train, x_test = train_test_split(data, test_size=.2)
"""
from sklearn.model_selection import train_test_split

# In order to randomly split our three data sets and to ensure our target
# variable (revenue) observations are paired with the correct predictor 
# variable observations, let's combine both into one data frame first.
df1 = pd.concat([y_rev1, X1], axis=1)
df2 = pd.concat([y_rev2, X2], axis=1)
df3 = pd.concat([y_rev3, X3], axis=1)
 
df1_train, df1_test = train_test_split(df1, test_size=.20)
df2_train, df2_test = train_test_split(df2, test_size=.20)
df3_train, df3_test = train_test_split(df3, test_size=.20)

# Re-train our models with DistToArena, GamesWatched, FanComplaints
# Let's do this in scikit-learn
mlr1_for_pred = LinearRegression()
mlr2_for_pred = LinearRegression()
mlr3_for_pred = LinearRegression()

mlr1_for_pred.fit(df1_train.iloc[:,1:], df1_train.iloc[:,0:1])
mlr2_for_pred.fit(df2_train.iloc[:,1:], df2_train.iloc[:,0:1])
mlr3_for_pred.fit(df3_train.iloc[:,1:], df3_train.iloc[:,0:1])

mlr1_preds = mlr1_for_pred.predict(df1_test.iloc[:,1:])
mlr2_preds = mlr2_for_pred.predict(df2_test.iloc[:,1:])
mlr3_preds = mlr3_for_pred.predict(df3_test.iloc[:,1:])

mse1 = mean_squared_error(df1_test.iloc[:,0:1], mlr1_preds)
mse2 = mean_squared_error(df2_test.iloc[:,0:1], mlr2_preds)
mse3 = mean_squared_error(df3_test.iloc[:,0:1], mlr3_preds)

r_sq1 = r2_score(df1_test.iloc[:,0:1], mlr1_preds)
r_sq2 = r2_score(df2_test.iloc[:,0:1], mlr2_preds)
r_sq3 = r2_score(df3_test.iloc[:,0:1], mlr3_preds)

print(mse1, mse2, mse3)
print(r_sq1, r_sq2, r_sq3) 
# R^2 for income level 1: 12.40%
# R^2 for income level 1: 47.84%
# R^2 for income level 1: 28.66%

