#################################################
#################################################
##          COPYRIGHT - DATA SOCIETY           ##
#################################################
#################################################

###################################
##    LINEAR REGRESSION, PT 2    ##
###################################

## NOTE: To run code blocks (separated by #%%), select somewhere inside the block 
##       and press ctrl + enter for Windows or Linux, 
##       and command + enter for Macs

#==================================================================================
#%%

#####

# Note: PLEASE RUN THE SCRIPT FROM LINEAR REGRESSION PT 1, AND
#       SET YOUR DIRECTORY TO THE FOLDER FROM LINEAR REGRESSION PT 1

#####

#==================================================================================
#%%
# Slide 9: Plot the residuals

# Create a new variable to represent residuals
slr_rev_income_stat_resid = slr_results.resid

# Plot the distribution of the residuals
plt.hist(slr_rev_income_stat_resid, 
         histtype = 'bar', 
         bins = 20, 
         ec='white', color='orange', 
         zorder=3)
plt.xlabel('Residuals', fontsize = 24)
plt.ylabel('Frequency', fontsize = 24)
plt.grid(.25, linestyle='dashed', zorder=0)

plt.savefig("lec2_slr_residuals.png", bbox_inches='tight')

#==================================================================================
#%%
# Slide 11: Calculate the variance of the residuals

# Calculate the variance of the residuals
# var() method produces the variance of the dataset
slr_rev_income_stat_resid.var(ddof = 1) # ddof = degrees of freedom

# Let's check manually to see if we understand what Python is doing.
slr_resid_avg = slr_rev_income_stat_resid.mean()
slr_resid_avg

# Subtract the average of the residuals from each residual
slr_resid_less_avg = slr_rev_income_stat_resid - slr_resid_avg

# Square this quantity, take the sum, divide by number of observations-1
variance_check = ((slr_resid_less_avg**2).sum()) / (len(data) - 1)
variance_check

#==================================================================================
#%%
# Slide 12: Standard deviation of residuals

slr_rev_income_stat_resid.std()

# Take the square root of the variance to make sure you know what Python is
# doing. Note that there are three ways to do this.
import math

slr_rev_income_stat_resid.var(ddof = 1)**(1/2)
math.sqrt(slr_rev_income_stat_resid.var(ddof = 1))
np.sqrt(slr_rev_income_stat_resid.var(ddof = 1))

#==================================================================================
#%%

############################
##    GO TO EXERCISE 1    ##
############################

#==================================================================================
#%%
### Slide 18: QQ Plot

# Import statsmodels
import statsmodels.api as sm

sm.qqplot(slr_rev_income_stat_resid, line='q')
plt.title('Q-Q Plot to Test Normality of Residuals', fontsize = 30)
plt.xlabel('Theoretical Quantiles', fontsize = 24)
plt.ylabel('Sample Quantiles', fontsize=24)
plt.savefig("lec2_slr_qqplot_w_line.png", bbox_inches='tight')
plt.grid(.25)

#==================================================================================
#%%
### Slide 21: Confidence interval of regression model

# The standard error, or confidence interval of a regression line, tells you
# with X% certainty where the best fit line should be.

# Plot regression line and 95% confidence intervals with sns.lmplot
sns.lmplot(x = 'Income', 
           y = 'TotRevSpend', 
           data = data, 
           line_kws = {'color': 'orange', 'alpha': 5}, 
           scatter_kws = {'color': 'black', 'alpha': .15}, 
           ci = 95)
plt.xlabel('Income ($)', fontsize = 24)
plt.ylabel('Total Revenue Spent ($)', fontsize = 24)

plt.savefig("slr_CI.png", bbox_inches = 'tight')

#==================================================================================
#%%

############################
##    GO TO EXERCISE 2    ##
############################

#==================================================================================
#%%
### Slide 29: Box-and-whisker plots

# Identify outliers with box-and-whisker plot.
outlier_data = np.column_stack(([1,2,3,4,5,6,7,8,9,10,20],
                                [2,4,3,2,6,8,10,6,7,9,20]))
outlier_df = pd.DataFrame(outlier_data, columns=['X values', 'Y values'])
bp = plt.boxplot(outlier_data) # fliersize is size of outlier points
plt.xticks([1, 2],['X values', 'Y values'])
plt.title('Data Set 1', fontsize=24)

plt.savefig("boxplot1.png", bbox_inches='tight')

# Obtain data from the box plot
outliers = [flier.get_ydata() for flier in bp['fliers']]
outliers

#==================================================================================
#%%
### Slide 30: Box-and-whisker plots

outlier_data2 = np.column_stack(([1,2,3,4,5,6,7,8,9,10,20],
                                 [2,4,3,2,6,8,10,7,6,9,5]))

outlier_df2 = pd.DataFrame(outlier_data2, columns=['X values', 'Y values'])
bp2 = plt.boxplot(outlier_data2)
plt.xticks([1,2], ['X values', 'Y values'])
plt.title('Data Set 2', fontsize=24)
plt.savefig("boxplot2.png", bbox_inches='tight')

outliers2 = [flier.get_ydata() for flier in bp2['fliers']]
outliers2

#==================================================================================
#%%
### Slide 32: Cook's distance

# Create a sample data set. 
outlier_data = np.column_stack(([1,2,3,4,5,6,7,8,9,10,20],
                                [2,4,3,2,6,8,10,6,7,9,20]))
outlier_df = pd.DataFrame(outlier_data, columns=['X values', 'Y values'])
#plt.scatter(outlier_df.iloc[:,0], outlier_df.iloc[:,1])

# Reshape data for input into model
X_rs = outlier_df.iloc[:,0].values.reshape((len(outlier_df),1))
Y_rs = outlier_df.iloc[:,1].values.reshape((len(outlier_df),1))
X = outlier_df.iloc[:,0]
Y = outlier_df.iloc[:, 1]

# Run a linear regression model
lm_outlier_data_stat = ols(formula='Y_rs ~ X_rs', data=outlier_df)
lm_outlier_fitted = lm_outlier_data_stat.fit()
influence = lm_outlier_fitted.get_influence()

# outlier_data_CD is the distance and p is the p-value
(outlier_data_CD, p) = influence.cooks_distance

#==================================================================================
#%%
### Slide 33: Cook's distance

plt.stem(np.arange(len(outlier_data_CD)), outlier_data_CD)

# A conventional cut-off point for Cook's distance is 4 / number of data points
# in the data set, all points where Cook's distance exceeds that value should
# be investigated as potential outliers
outlier_data_CD_select = outlier_data_CD[outlier_data_CD>4/11]
outlier_data_CD_select

# Visualize Cook’s distance
plt.stem(np.arange(len(outlier_data_CD)), outlier_data_CD)
plt.xlabel('Observations', fontsize=18)
plt.ylabel('Cook Distance', fontsize=18)

#==================================================================================
#%%
### Slide 34: Cook's distance


print(pd.DataFrame(outlier_data_CD, columns=['Cook Distance']))

outlier_data_CD_select

#==================================================================================
#%%
### Slide 34: Outliers: NBA Fan Behavior

# Let's see if our total fan revenue model comparing total revenue spent and
# fan income level has any outliers. 
# Let's start with a boxplot. 
# Matplotlib box plots can only take lists, tuples or numpy arrays
# not data frames.

rev_income_arr = np.array(data[['TotRevSpend','Income']])
plt.boxplot(rev_income_arr[:,0:1])
plt.xticks([1],['Total Revenue Spent']) # No outliers

plt.boxplot(rev_income_arr[:, 1:2]) # No outliers
plt.xticks([1], ['Income'])

#==================================================================================
#%%
### Slide 37: Cook's distance: view outliers

# Let's try Cook's Distance
# Remember our SLR model
slr_results_influence = slr_results.get_influence()

# slr_rev_income_CD is the distance and p is the p-value
(slr_rev_income_CD, p) = slr_results_influence.cooks_distance
plt.stem(np.arange(len(slr_rev_income_CD)), slr_rev_income_CD)
plt.title('Cook Distance', fontsize = 24)
plt.grid(.25)

slr_rev_income_CD_select = slr_rev_income_CD[slr_rev_income_CD > 4 / len(data)]
data_CD_influencers = data[slr_rev_income_CD > 4 / len(data)]

#==================================================================================
#%%
### Slide 38: Cook's distance: view outliers

# There are 3 potential outliers: observations 4, 50, 621    
sns.lmplot(x='Income', y='TotRevSpend', data = data, 
           line_kws = {'color': 'orange'}, 
           scatter_kws = {'color': 'black'})
plt.scatter(data_CD_influencers['Income'], 
            data_CD_influencers['TotRevSpend'], 
            color = 'red')
plt.xlabel('Income ($)', fontsize = 18)
plt.ylabel('Total Revenue Spent ($)', fontsize = 18) 

   
#==================================================================================
#%%
### Slide 40: Cook's distance: view outliers

### Cook's Distance: remove outliers
# Observation 621 seems to stick with the general pattern of the data and
# the boxplot didn't pick it up as an outlier, so let's not drop it from 
# the data. Given incomes of around $35,000, it does seem rather out of the 
# ordinary for these fans to spend $900 on NBA entertainment which in 
# general range from spending $0 to about $600. However, it is totally possible
# that these lower income fans just choose to allocate their wealth in this
# way. 

# let's remove observations 4 and 50 and see how the analysis is affected. 
data_clean = data.drop(data.index[[4,50]])
y_rev_inc_clean = data_clean['TotRevSpend']
X_rev_inc_clean = data_clean['Income']
slr_rev_income_stat_clean = ols(formula='y_rev_inc_clean ~ X_rev_inc_clean', 
                                data=data_clean)
slr_results_clean = slr_rev_income_stat_clean.fit()
slr_results_clean.summary() 
# slope = 0.0022, intercept= 246.6058
# Adjusted R^2 = 0.144

# Original model
slr_results.summary()
# slope = 0.0021, intercept = 251.3737
# Adjusted R^2 = 0.139

# These two outliers don't make much of a difference in the analysis, 
# just a change in intercept.

#==================================================================================
#%%
### Slide 42: Regression without outliers

# Now let's plot the data without the outliers
sns.lmplot(x = 'Income', y = 'TotRevSpend', 
           data=data_clean, 
           line_kws = {'color': 'orange'}, 
           scatter_kws = {'color': 'black'})
plt.xlabel('Income ($)', fontsize = 18)
plt.ylabel('Total Revenue Spent ($)', fontsize = 18)

# I would argue against removing these two observations from the data
# since the regression results didn't change much and these observations 
# are believable based on how die-hard the fan is. 

#==================================================================================
#%%
### Slide 45: Multiple linear regression with and outliers

# Let's consider our final model from lecutre 1:
# DistToArena, GamesWatched, Income, FanComplaints
X_final2 = data[['DistToArena', 'GamesWatched', 'Income', 'YrsInDatabase']]
y_final2 = data['TotRevSpend']
mlr_final2_stat = ols(formula = 'y_final2 ~ X_final2', data = data)
mlr_final2_results = mlr_final2_stat.fit()
mlr_final2_results.summary()

#==================================================================================
#%%
### Slide 47: Multiple linear regression with and outliers

mlr_results_influence = mlr_final2_results.get_influence()

# mlr_final2_CD is the distance and p is the p-value
(mlr_final2_CD, p) = mlr_results_influence.cooks_distance
plt.stem(np.arange(len(mlr_final2_CD)), mlr_final2_CD)
plt.title('Cook Distance', fontsize=24)
plt.grid(.25)

mlr_final2_CD_select = mlr_final2_CD[mlr_final2_CD > 4 / len(data)]
data2_CD_influencers = data[mlr_final2_CD > 4 / len(data)]

#==================================================================================
#%%
### Slide 49: Multiple linear regression with and outliers

### NOTE - CODE FOR SLIDES NOT SHOWN IN SLIDES. 

# According to Cook's distance and our threshold of 0.004, there are 24
# possible outliers. 
# Let's plot each 
# 24 outliers - 4/1000 threshold

fig = plt.figure()
ax1 = fig.add_subplot(231)
ax1.hist(data2_CD_influencers['TotRevSpend'], 
         bins = 20, ec = 'black')
ax1.set_title('Total Revenue Spent')

ax2 = fig.add_subplot(232)
ax2.hist(data2_CD_influencers['DistToArena'], 
         bins = 20, ec = 'black')
ax2.set_title('Distance to Arena')

ax3 = fig.add_subplot(233)
ax3.hist(data2_CD_influencers['GamesWatched'], 
         bins = 20, ec = 'black')
ax3.set_title('Games Watched')

ax4 = fig.add_subplot(234)
ax4.hist(data2_CD_influencers['Income'], 
         bins = 20, ec = 'black')
ax4.set_title('Income')

ax5 = fig.add_subplot(235)
ax5.hist(data2_CD_influencers['YrsInDatabase'], 
         bins = 20, ec = 'black')
ax5.set_title('YrsInDatabase')

#==================================================================================
#%%
### Slide 50: Multiple regression: outliers

# It looks like many of these "outliers" live relatiely close to the arena,
# watched about 30 games, have higher incomes and have been in the database
# for a number of years. These just seem like satisfied customers!

# Let's try to limit the selected outliers to ensure that
# we don't exclude any points that may not be outliers. Instead of 4 in the
# numerator, we'll double it. 
data2_CD_influencers2 = data[mlr_final2_CD > (4*2) / len(data)]

# View the data identified by cook's method as potential outliers. 
# Select only the rows and columns we're interetsed in. 
data2_CD_influencers2 # observations 52, 54, 61, 72, 77, 92


#==================================================================================
#%%
### Slides 51-52: Multiple regression: outliers

# 6 outliers - 8/1000 threshold
outlier_obs = []
for i in range(len(data2_CD_influencers2)):
    outlier_obs.append(data2_CD_influencers2.index[i])
outlier_obs    

fig = plt.figure()
ax1 = fig.add_subplot(231)
ax1.stem(outlier_obs, data2_CD_influencers2['TotRevSpend'])
ax1.set_xticks(outlier_obs)
ax1.set_title('Total Revenue Spent')

ax2 = fig.add_subplot(232)
ax2.stem(outlier_obs, data2_CD_influencers2['DistToArena'])
ax2.set_xticks(outlier_obs)
ax2.set_title('Distance to Arena')

ax3 = fig.add_subplot(233)
ax3.stem(outlier_obs, data2_CD_influencers2['GamesWatched'])
ax3.set_xticks(outlier_obs)
ax3.set_title('Games Watched')

ax4 = fig.add_subplot(234)
ax4.stem(outlier_obs, data2_CD_influencers2['Income'])
ax4.set_xticks(outlier_obs)
ax4.set_title('Income')

ax5 = fig.add_subplot(235)
ax5.stem(outlier_obs, data2_CD_influencers2['YrsInDatabase'])
ax5.set_xticks(outlier_obs)
ax5.set_title('YrsInDatabase')

#==================================================================================
#%%
### Slide 54: Multiple regression: outliers

# It looks like these 6 observations make more than $100,000 in income, have
# mostly spent at least 15 years in the database, and watched about 30 games.
# Observation 92 seems to be a newer NBA fan having watched only a few games
# but in the games the fan watched he/she spent a lot of money.  

# If we remove these observations let's see how the regression analysis might
# change. 
data2_clean = data.drop(data.index[outlier_obs])
X_final2_clean = data2_clean[['DistToArena', 'GamesWatched', 
                              'Income', 'YrsInDatabase']]
y_final2_clean = data2_clean['TotRevSpend']
mlr_final2_stat_clean = ols(formula='y_final2_clean ~ X_final2_clean', 
                            data = data2_clean)
mlr_final2_results_clean = mlr_final2_stat_clean.fit()
mlr_final2_results_clean.summary()

#%%

# Original model (all coefficients signifincant, model significant)
# Adjusted R^2 = 0.557
# Intercept = 285.4069
# DistToArena = -0.5334
# GamesWatched = 6.6455
# Income = 0.0011
# YrsInDatabase = 18.3647

# Model without outliers (all coefficients significant, model significant)
# Adjusted R^2 = 0.571 (2.51% increase)
# Intercept = 283.6414 (0.62% decrease)
# DistToArena = -0.5443 (2.04% decrease)
# GamesWatched = 6.7674 (1.83% increase)
# Income = 0.0011 (negligable change)
# YrsInDatabase = 19.2615 (4.88% increase)

# While most of the coefficients did change at least a little, the conclusions
# remain the same!

#==================================================================================
#%%
### Slide 60: Validating: check for multicollinearity

### Validating: check for multicollinearity
# We will use the variance_inflation_factor function from statsmodels package
# to check for multicollinearity.
from statsmodels.stats.outliers_influence import variance_inflation_factor

# Consider a model including all 6 predictor variables
X_all_multicollinearity = [variance_inflation_factor(data.iloc[:,1:].values, j) 
    for j in range(data.iloc[:,1:].shape[1])]
X_all_multicollinearity

# While none of the variables have egregious VIFs there is some in all and 
# VIFs certainly over 5 are worth looking into. 
data.iloc[:, 1:].corr()

#==================================================================================
#%%
### Slide 62: Validating: check for multicollinearity

# Since FanComplaints and FanSatisfaction are strongly negatively correlated,
# let's run a model without FanSatisfaction and examine its predictive power.
X_no_FanSatisfaction = data.iloc[:,1:].drop('FanSatisfaction', 
                                axis = 1)
X_no_FanSatisfaction_multicollinearity = [variance_inflation_factor(
        X_no_FanSatisfaction.values, i)
        for i in range(X_no_FanSatisfaction.shape[1])]
X_no_FanSatisfaction_multicollinearity
# The VIFs decreased across the board to all below 5. 

pd.concat([pd.DataFrame(X_no_FanSatisfaction.columns, 
                        columns = ['Variables']),
           pd.DataFrame(X_no_FanSatisfaction_multicollinearity, 
                        columns = ['VIF'])],
           axis = 1)

#==================================================================================
#%%
### Slide 63: Validating: check for multicollinearity

# Let's consider our final model from Lecture 1 which included:
# DistToArena, GamesWatched, Income, FanComplaints
X_final2_multicollinearity = [variance_inflation_factor(
                                X_final2.values, i) 
                                for i in range(X_final2.shape[1])]
X_final2_multicollinearity
# All VIFs under 5, so we're getting better. 

# What if we standardize our data?
X_final2_scaled = preprocessing.scale(X_final2)
X_final2_scaled_multicollinearity = [variance_inflation_factor(
        X_final2_scaled, i) for i in range(X_final2_scaled.shape[1])]
X_final2_scaled_multicollinearity

#==================================================================================
#%%
### Slide 66: Validating: check the residuals

# Let’s consider the final model from lecture 1 with standardized data. This model 
# includes: DistToArena, Income, GamesWatched, YrsInDatabase. 

plt.hist(mlr_final_results.resid, 
         bins = 20, 
         color = 'orange', 
         ec = 'white',
         zorder = 3)
plt.xlabel('MLR Final Model Residuals', 
           fontsize = 24)
plt.grid(.25, 
         linestyle = 'dashed', 
         zorder = 0)

# The residuals seem normal, let's check the Q-Q-plot!

#==================================================================================
#%%
### Slide 68: Validating: check the residuals

# Now let's check the residuals for normality of distribution using the 
# Q-Q plot
sm.qqplot(mlr_final_results.resid, line='s')
plt.title('Q-Q Plot to Test Normality of Residuals', fontsize=24)
plt.xlabel('Theoretical Quantiles', fontsize=18)
plt.ylabel('Sample Quantiles', fontsize=18)
plt.grid(.25)
# Our residuals look normally distributed!

#==================================================================================
#%%
### Slide 69: Validating: check the residuals

# Does the variance of the residuals change with the predicted value?
plt.scatter(mlr_final_results.predict(), 
            mlr_final_results.resid, 
            color = 'black')
plt.plot(mlr_final_results.predict(), 
         np.zeros((1000,)), 
         color = 'orange')
plt.xlabel('Fitted Value', 
           fontsize = 18)
plt.ylabel('Residual', 
           fontsize = 18)


# Variance of residuals looks like it increases until the fitted values hit 0
# and then decreases after (with a bunch of exceptions at the ends).

plt.scatter(mlr_final_results.predict(), mlr_final_results.resid.abs(), 
            color = 'black')
plt.plot(mlr_final_results.predict(), np.zeros((1000,)), 
         color = 'orange')
plt.xlabel('Fitted Value', fontsize = 18)
plt.ylabel('Absolute Value of Residual', fontsize = 18)

# There should not be a pattern but there is clearly a triangular shape
# to the absolute value of the residuals. This means we're not taking something
# into account.

#==================================================================================
#%%
### Slide 74: Homoscedasticity example

# Example with homoscedasticity (no heterscedasticity present)
sample_data = np.column_stack(([1,2,3,4,5,6,7,8,9,10,20],
                               [2,4,3,2,6,8,10,7,6,9,20]))

sample_df = pd.DataFrame(sample_data, columns=['x_data', 'y_data'])
sample_lm = ols('y_data ~ x_data', data=sample_df).fit()
plt.scatter(sample_df['x_data'], sample_df['y_data'], color='black')
plt.plot(sample_df['x_data'], sample_lm.predict(), color='orange')

#==================================================================================
#%%
### Slide 75: Homoscedasticity example

# Plot residuals vs. fitted values.
plt.scatter(sample_lm.predict(), sample_lm.resid, color='black')
plt.plot(sample_lm.predict(), np.zeros((11,)), color='orange')
plt.xlabel('Fitted Values', fontsize=18)
plt.ylabel('Residual', fontsize=18)
plt.grid(.25, 
         linestyle='dashed')

#==================================================================================
#%%
### Slide 76: Homoscedasticity example

### Breusch-Pagan test
# You can run the Breusch-Pagan test with het_breuschpagan function in the
# statsmodels.stats.api module. 
import statsmodels.stats.api as sms
from statsmodels.compat import lzip


# Labels for output of test
name = ['Lagrange multiplier statistic', 'p-value',
        'f-value', 'f p-value'] # names of output from test
test_sample = sms.het_breuschpagan(sample_lm.resid, sample_lm.model.exog)
lzip(name, test_sample) # pairs each output from test with its meaning

#==================================================================================
#%%
### Slide 77: Homoscedasticity example

# Both p-values are very large for the sample data, which means that the
# regression model for the error terms has poor explanatory power, which in 
# turn means that there is NO heteroscedasticity.

test_mlr = sms.het_breuschpagan(mlr_final_results.resid, 
                                mlr_final_results.model.exog)                           
lzip(name, test_mlr)

# While the p-values are not as high as the sample data, they are still above
# our 5% significance level. This means that our regression model for the 
# error terms has some explanatory power, but NOT enough to say that there is
# heteroscedasticity present.

# If you would rather trus the residual plot rather than the Breusch-Pagan
# test, there are various ways to deal with non-constant variance, including:
    # consider transforming the dependent or independent variables
    # weighted least squares
    # add polynomial terms or interaction terms
    
#==================================================================================
#%%

############################
##    GO TO EXERCISE 4    ##
############################
    
#==================================================================================
#%%
### Slide 83: Variable interactions

# You can test the interaction of different factors in the model using the
# "*" symbol.
# First let's split up our income variables as we did in the exercises for 
# lecture 1.
data.loc[data['Income']<=50000, 'IncomeLevel'] = 1
data.loc[(data['Income']>50000) & (data['Income']<=125000), 
         'IncomeLevel'] = 2
data.loc[data['Income']>125000, 'IncomeLevel'] = 3         

# Build interaction model
rev = data['TotRevSpend']
gw = data['GamesWatched']
il = data['IncomeLevel']
rev_il_gw_interaction_lm = ols(
        formula='rev ~ gw*il', 
        data = data).fit() 

# If you want il encoded with dummy variables, use C(il) in the formula. 
rev_il_gw_interaction_lm.summary()   

#==================================================================================
#%%
### Slide 85: Variable interactions

# To do this in scikit-learn:
# Not one hot encoded
poly = preprocessing.PolynomialFeatures(2, interaction_only=True, 
                                        include_bias=False)    
il_gw_interaction_sklearn = poly.fit_transform(data[['GamesWatched', 
                                                     'IncomeLevel']])

rev_il_gw_interaction_lm_sklearn = LinearRegression()
rev_il_gw_interaction_lm_sklearn.fit(il_gw_interaction_sklearn, rev)
rev_il_gw_interaction_fitted = rev_il_gw_interaction_lm_sklearn.predict(il_gw_interaction_sklearn)

r2_score(rev, rev_il_gw_interaction_fitted)

#==================================================================================
#%%
### Slide 86: Variable interactions

# One hot encoded
from sklearn.preprocessing import OneHotEncoder
integer_encoded = data['IncomeLevel'].values.reshape(len(data['Income']), 1)
onehot_encoder = OneHotEncoder(sparse=False)
onehot_encoded = onehot_encoder.fit_transform(integer_encoded)
poly_onehot = preprocessing.PolynomialFeatures(2, interaction_only=True, include_bias=False)
il_gw_interaction_lm_sklearn_onehot = poly_onehot.fit_transform(data['GamesWatched'].values.reshape(-1,1), onehot_encoded)
rev_il_gw_interaction_lm_sklearn_onehot = LinearRegression()
rev_il_gw_interaction_lm_sklearn_onehot.fit(il_gw_interaction_lm_sklearn_onehot, rev)
rev_il_gw_interaction_onehot_fitted = rev_il_gw_interaction_lm_sklearn_onehot.predict(il_gw_interaction_lm_sklearn_onehot)
r2_score(rev, rev_il_gw_interaction_onehot_fitted)

#==================================================================================
#%%
### Slide 87: Variable interactions

### Let's include interaction terms in our model selection process 
# from lecture 1.
rev_scaled = y_scaled
dist_scaled = X_scaled[:,0:1]
gw_scaled = X_scaled[:,1:2]
inc_scaled = X_scaled[:,2:3]
fs_scaled = X_scaled[:,3:4]
yid_scaled = X_scaled[:,4:5]
fc_scaled = X_scaled[:, 5:6]
mlr_interaction = ols(formula='rev_scaled ~ dist_scaled + gw_scaled + inc_scaled + fs_scaled + yid_scaled + fc_scaled + gw_scaled*il + fs_scaled*il',
                      data=pd.DataFrame(data_scaled)).fit()
mlr_interaction.summary()

#==================================================================================
#%%
### Slide 89: Variable interactions

# To use RFE for model selection we need a LinearRegression object of sklearn
mlr_poly = preprocessing.PolynomialFeatures(2,
                                        interaction_only=True,
                                        include_bias=False)

gw_scaled_il_to_transform = np.column_stack([gw_scaled, data['IncomeLevel'].values])

gw_scaled_il_interaction = mlr_poly.fit_transform(gw_scaled_il_to_transform)

# Add the il and interaction column to our already scaled matrix of predictors
X_scaled_interaction = np.column_stack([X_scaled, gw_scaled_il_interaction[:,1:]])

# drop income column now that we have 

mlr_interaction_sklearn = LinearRegression()
mlr_interaction_sklearn.fit(X_scaled_interaction, y_scaled)

#==================================================================================
#%%
### Slide 90: Variable interactions

R_squared_comparison = []

for i in range(1, X_scaled_interaction.shape[1]+1):
    rfe = RFE(mlr_interaction_sklearn, i)
    rfe_fit = rfe.fit(X_scaled_interaction, np.ravel(y_scaled))
    X = X_scaled_interaction[:,np.ravel(np.where(rfe_fit.ranking_==1))]
    mlr = LinearRegression()
    mlr.fit(X, y_scaled)
    y_mlr_fitted = mlr.predict(X)
    R_sq = r2_score(y_scaled, y_mlr_fitted)
    R_squared_comparison.append(R_sq)

plt.plot(range(1, X_scaled_interaction.shape[1]+1), R_squared_comparison)
plt.title('Model Comparison Using RFE With Interaction', fontsize=24)
plt.xlabel('Number of Most Important Features', fontsize=18)
plt.ylabel('R^2', fontsize=18)
plt.grid(1)

#==================================================================================
#%%
### Slide 92: Variable interactions

# Do any of the interaction terms show up in our most important features?

# Similar to lecture 1, 4 features seems to be the optimal number without too much
# model complexity.

rfe_interaction = RFE(mlr_interaction_sklearn, 4)
rfe_interaction_fit = rfe_interaction.fit(X_scaled_interaction, np.ravel(y_scaled))
print(rfe_interaction_fit.ranking_)


# Our best model according to RFE even with the possibility of an interaction
# between games watched and income level did not change. 

# Perhaps the relationships in the real world are simply non-linear?

#==================================================================================
#%%
### Slide 93: Utilize the model for prediction

# If you decide to use this model and need to make a prediction based on 
# some input variables, here is how you would do that:
test = xls.parse('Prediction')    
del test['Unnamed: 0']

# Standardize test data
new_data = preprocessing.scale(test.iloc[:,2:])

# select DistToArena, GamesWatched, Income, YrsInDatabase
# delete other variables
X_scaled_for_pred = np.delete(X_scaled, [3,5], axis=1)
new_data_for_pred = np.delete(new_data, [3,5], axis=1) 

# First fit the model to the data then predict revenue spent with the new data.
mlr = LinearRegression()
mlr.fit(X_scaled_for_pred, y_scaled)
mlr_prediction = mlr.predict(new_data_for_pred)
print(mlr_prediction)

# These are your predictions for the new data! Remember that these are standardized predictions!

#==================================================================================
#%%

############################
##    GO TO EXERCISE 5    ##
############################
    

